import datetime
import json
import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type

from thefuzz import process  # type: ignore

import cf2tf.conversion.expressions as functions
import cf2tf.terraform._configuration as config
import cf2tf.terraform.doc_file as doc_file
from cf2tf.conversion.overrides import GLOBAL_OVERRIDES, OVERRIDE_DISPATCH
from cf2tf.terraform.blocks import Block, Locals, Output, Resource, Variable
from cf2tf.terraform.hcl2 import AllTypes
from cf2tf.terraform.hcl2.complex import ListType, MapType
from cf2tf.terraform.hcl2.custom import CommentType, LiteralType
from cf2tf.terraform.hcl2.primitive import (
    BooleanType,
    NumberType,
    StringType,
    TerraformType,
)

if TYPE_CHECKING:
    from cf2tf.terraform.code import SearchManager


# Wish I could go back to this style of import but getting cycle error
# from cf2tf.terraform import doc_file, Configuration

ResourceName = str
ResourceValues = Dict[str, Any]
Template = Dict[str, Any]
CFResource = Tuple[ResourceName, ResourceValues]
CFResources = List[CFResource]
Manifest = Dict[str, CFResources]

CFDict = Dict[str, Dict[str, Any]]

log = logging.getLogger("cf2tf")


class TemplateConverter:
    def __init__(
        self, template_name: str, cf_template: CFDict, search_manager: "SearchManager"
    ) -> None:
        self.name = template_name
        self.cf_template = cf_template
        self.search_manager = search_manager
        self.terraform = config.Configuration([])
        self.all_resources: Optional[CFResources]
        self.manifest: Manifest = {}

        self.post_proccess_blocks: List[Block] = []

        # This is not only the sections we are interested in, but the conversion order,
        # which is also important.
        self.valid_sections = [
            "Parameters",
            "Mappings",
            "Conditions",
            "Resources",
            "Outputs",
        ]

    @staticmethod
    def _json_encoder(value: Any) -> Any:
        if isinstance(value, (datetime.date, datetime.datetime)):
            return value.isoformat()
        else:
            return value

    def add_post_block(self, block: Block):
        block_refs = [block.base_ref() for block in self.post_proccess_blocks]

        log.debug(block_refs)

        if block.base_ref() not in block_refs:
            self.post_proccess_blocks.insert(0, block)

    def get_block_by_type(self, block_type: Type[Block]):
        for block in self.post_proccess_blocks:
            if isinstance(block, block_type):
                return block

    def convert(self) -> config.Configuration:
        # Should convert the given cloudformation template to a terraform configuration
        self.parse_template()

        # This is used by the resolve values function
        self.all_resources = [
            resource for _, resources in self.manifest.items() for resource in resources
        ]

        # These are the resources converted directly from cloudformation
        tf_resources = self.convert_to_tf(self.manifest)

        # We also dynamicly generate new resources like the locals block
        tf_resources[:0] = self.post_proccess_blocks

        return config.Configuration(tf_resources)

    def parse_template(self):
        for section in self.valid_sections:
            if section not in self.cf_template:
                log.debug(
                    f"Ignoring section {section} not found in {self.valid_sections}"
                )
                continue

            section_values = self.cf_template[section]

            self.manifest[section] = list(section_values.items())

        log.debug(
            f"Parsed the following resources for processing:\n{json.dumps(self.manifest, default=self._json_encoder)}"
        )

    def resource_lookup(
        self, resource_name: str, sections: List[str]
    ) -> Optional[Dict[str, Any]]:
        for section in sections:
            section_resources = self.manifest.get(section)

            if section_resources:
                section_map: Dict[str, Dict[str, Any]] = dict(section_resources)  # type: ignore

                if resource_name in section_map:
                    resource = section_map[resource_name]
                    return resource

        return None

    def convert_to_tf(self, manifest: Manifest):
        tf_resources: List[Block] = []

        for section in list(manifest):
            resources = manifest[section]
            log.debug(f"Converting {len(resources)} {section}")

            converter = get_converter(self, section)

            tf_resources.extend(converter(resources))

        return tf_resources

    def resolve_values(  # noqa: max-complexity=13
        self,
        data: Any,
        allowed_func: functions.Dispatch,
        prev_func: Optional[str] = None,
    ):
        """Recurses through a Cloudformation template. Solving all
        references and variables along the way.

        Args:
            data (Any): Could be a dict, list, str or int.

        Returns:
            Any: Return the rendered data structure.
        """

        if isinstance(data, dict):
            key: str

            for key in list(data):
                value = data[key]

                if key == "Ref":
                    return functions.ref(self, value)

                # This takes care of keys that not intrinsic functions,
                #  except for the condition func
                if "Fn::" not in key and key != "Condition":
                    data[key] = self.resolve_values(value, allowed_func, prev_func)
                    continue

                # Takes care of the tricky 'Condition' key
                if key == "Condition":
                    # The real fix is to not resolve every key/value in the entire
                    # cloudformation template. We should only attempt to resolve what is needed,
                    # like outputs and resource properties.
                    # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference.html

                    if "Properties" in data or "Value" in data:
                        continue

                    # If it's an intrinsic func
                    if is_condition_func(value):
                        return functions.condition(self, value)

                    # Normal key like in an IAM role
                    data[key] = self.resolve_values(value, allowed_func, prev_func)
                    continue

                if key not in allowed_func:
                    raise ValueError(f"{key} not allowed to be nested in {prev_func}.")

                prev_func = key

                value = self.resolve_values(
                    value, functions.ALLOWED_FUNCTIONS[key], prev_func
                )

                try:
                    return allowed_func[key](self, value)
                except Exception as e:
                    return CommentType(
                        f"Unable to resolve {key} with value: {value} because {e}"
                    )

            return MapType(data)
        elif isinstance(data, list):
            resolved_list_values = [
                self.resolve_values(item, allowed_func, prev_func) for item in data
            ]

            return ListType(resolved_list_values)
        elif isinstance(data, bool):
            return BooleanType(data)
        elif isinstance(data, str):
            return StringType(data)
        elif isinstance(data, (int, float)):
            return NumberType(data)
        elif isinstance(data, datetime.date):
            return StringType(str(data))
        else:
            log.error(f"Found type {type(data)} with value {data}")
            raise Exception(f"Unknown value {data}, in resolve function.")

    def convert_parameters(self, parameters: CFResources):
        tf_vars: List[Variable] = []

        for param_name, param_value in parameters:
            log.debug(
                f"Converting Cloudformation Parameter - {param_name} to Terraform."
            )

            tf_name = pascal_to_snake(param_name)
            log.debug(f"Converted name to {tf_name}")

            tmp_props: Dict[str, Any] = {
                "description": param_value.get("Description", ""),
                "type": param_value.get("Type", "").lower(),
                "default": param_value.get("Default", ""),
            }

            converted_arguments = {k: v for k, v in tmp_props.items() if v != ""}

            resolved_args = self.resolve_values(converted_arguments, {})

            if "type" in resolved_args:
                resolved_args["type"] = convert_parameter_type(resolved_args["type"])

            var = Variable(tf_name, MapType(resolved_args))

            tf_vars.append(var)

            # Add space for easier to read logging output
            add_space()

        return tf_vars

    def convert_mappings(self, mappings: CFResources):
        log.debug("Converting Mappings to Terraform Locals block.")

        dict_mappings = dict(mappings)

        resolved_values = self.resolve_values(dict_mappings, functions.ALL_FUNCTIONS)

        mapping_args = {"mappings": resolved_values}

        local_block = Locals(mapping_args)

        self.post_proccess_blocks.append(local_block)

        return []

    def convert_conditions(self, conditions: CFResources):
        log.debug("Converting Conditions to Terraform Locals block.")

        dict_conditons = dict(conditions)

        resolved_values = self.resolve_values(dict_conditons, functions.ALL_FUNCTIONS)

        log.debug(resolved_values)

        local_block = self.get_block_by_type(Locals)

        if local_block:
            local_block.arguments.update(resolved_values)
            return []

        local_block = Locals(resolved_values)

        self.post_proccess_blocks.append(local_block)

        return []

    def convert_resources(self, resources: CFResources):
        tf_resources: List[Resource] = []

        for resource_id, resource_values in resources:
            log.debug(f"Converting Cloudformation resource {resource_id} to Terraform.")

            tf_name = pascal_to_snake(resource_id)
            log.debug(f"Converted name to {tf_name}")

            resource_type = resource_values.get("Type")

            if not resource_type:
                raise Exception("Type is required")

            docs_path = self.search_manager.find(resource_type)

            log.debug(f"Found documentation file {docs_path}")

            tf_type = create_resource_type(docs_path)

            log.debug(f"Converted type from {resource_values.get('Type')} to {tf_type}")

            valid_arguments, valid_attributes = doc_file.parse_attributes(docs_path)

            log.debug(
                f"Parsed the following arguments from the documentation: \n{valid_arguments}"
            )

            log.debug(
                f"Parsed the following attributes from the documentation: \n{valid_attributes}"
            )

            properties: Dict[str, Any] = resource_values.get("Properties", {})

            arguments = MapType(properties)

            if properties:
                log.debug(
                    "Converting the intrinsic functions to Terraform expressions..."
                )

                resolved_values = self.resolve_values(
                    properties, functions.ALL_FUNCTIONS
                )

                log.debug("Overiding Properties")

                overrided_values = perform_resource_overrides(
                    tf_type, resolved_values, self
                )

                overrided_values = perform_global_overrides(
                    tf_type, overrided_values, self
                )

                log.debug("Converting property names to argument names...")

                arguments = props_to_args(overrided_values, valid_arguments, docs_path)

                log.debug(f"Converted properties to {arguments}")

            conditional = resource_values.get("Condition")

            if conditional is not None:
                condition_map = {"count": LiteralType(f"local.{conditional} ? 1 : 0")}
                arguments = MapType({**condition_map, **arguments})

            resource = Resource(
                tf_name, tf_type, arguments, valid_arguments, valid_attributes
            )
            tf_resources.append(resource)

            # Add space for easier to read logging output
            add_space()

        return tf_resources

    def convert_outputs(self, outputs: CFResources):
        tf_outputs: List[Output] = []

        for output_id, output_props in outputs:
            log.debug(f"Converting Cloudformation Output {output_id} to Terraform.")

            tf_name = pascal_to_snake(output_id)
            log.debug(f"Converted name to {tf_name}")

            tmp_props: Dict[str, Any] = {
                "description": output_props.get("Description", ""),
                "value": output_props.get("Value"),
            }

            converted_args = {k: v for k, v in tmp_props.items() if v != ""}

            resolved_args = self.resolve_values(converted_args, functions.ALL_FUNCTIONS)

            log.debug(f"Converted properties to {resolved_args}")

            tf_outputs.append(Output(tf_name, resolved_args))

            # Add space for easier to read logging output
            add_space()

        return tf_outputs


def get_converter(
    item: Any,
    section_name: str,
) -> Callable[[CFResources], List[Block]]:
    conveter_name = f"convert_{section_name.lower()}"

    converter = getattr(item, conveter_name)
    return converter


def find_section(tf_attribute_name: str, docs_path: Path):
    """Checks to see if the attribute is also a subsection in the terraform documentation

    Args:
        tf_attribute_name (str): A terraform attribute name.

    Returns:
        str: The name of the section if found, otherwise none.
    """

    log.debug(f"Checking if {tf_attribute_name} has a section in {docs_path}.")

    # Search works better if we split the words apart, but we have to put it back together later
    search_term = tf_attribute_name.replace("_", " ")

    search_items = doc_file.all_sections(docs_path)

    # Score needs to bascially be an exact match
    result = matcher(search_term, search_items, 95)

    if not result:
        log.debug(f"{tf_attribute_name} does not have a section in {docs_path}")
        return ""

    result_name, ranking = result

    log.debug(f"Found section {result_name} with {ranking}% match.")

    return result_name


def pascal_to_snake(name: str):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def matcher(search_term: str, search_items: List[str], score_cutoff=0):
    result: Optional[Tuple[str, int]] = process.extractOne(
        # search_term, search_items, scorer=fuzz.token_sort_ratio
        search_term,
        search_items,
        score_cutoff=score_cutoff,
    )

    return result


def camel_case_split(text: str) -> str:
    items = re.findall(r"[A-Z\d](?:[a-z]+|\d|[A-Z]*(?=[A-Z]|$))", text)

    return " ".join(items) if items else text


def create_resource_type(doc_path: Path):
    file_base_name = doc_path.name.split(".")[0]
    return f"aws_{file_base_name}"


# todo Not used yet, but I think it will be in a future refactoring
def contains_functions(self, data: Dict[str, Any]):
    functions = ["Ref", "Fn::"]

    for key in list(data):
        if key in functions:
            return True

    return False


def props_to_args(
    cf_props: Dict[str, AllTypes],
    valid_tf_arguments: List[str],
    docs_path: Path,
):
    # Search works better if we split the words apart, but we have to put it back together later
    search_items = [item.replace("_", " ") for item in valid_tf_arguments]

    converted_attrs: Dict[str, AllTypes] = {}

    for prop_name, prop_value in cf_props.items():
        tf_arg_name, tf_arg_value = convert_prop_to_arg(
            prop_name, prop_value, search_items, docs_path
        )

        converted_attrs[tf_arg_name] = tf_arg_value

    log.debug(f"Converted {cf_props.keys()} to {converted_attrs.keys()}")

    return MapType(converted_attrs)


def convert_prop_to_arg(
    prop_name: str, prop_value: AllTypes, search_items: List[str], docs_path: Path
) -> Tuple[str, AllTypes]:
    search_term = camel_case_split(prop_name)

    log.debug(f"Searching for {search_term} instead of {prop_name}")

    result = matcher(search_term, search_items, 80)

    if not result:
        log.debug(f"No match found for {prop_name}, commenting out this argument.")
        # return f"// CF Property({prop_name})", str(prop_value)
        return prop_name, CommentType(f"CF Property({prop_name}) = {prop_value}")

    attribute_match, ranking = result

    # Putting the underscore back in
    tf_arg_name = attribute_match.replace(" ", "_")

    log.debug(f"Converted {prop_name} to {tf_arg_name} with {ranking}% match.")

    # Terraform sometimes has nested blocks, if prop_value is a map/list, its possible
    # that tf_attribute_name is a nested block in terraform

    if not isinstance(prop_value, (list, dict)):
        return tf_arg_name, prop_value

    try:
        tf_arg, tf_values = parse_subsection(tf_arg_name, prop_value, docs_path)
        return tf_arg, tf_values
    except Exception:
        raise Exception(
            f"Failed to parse subsection for {prop_name}/{tf_arg_name} in {docs_path}"
        )


def parse_subsection(
    arg_name: str, prop_value: AllTypes, docs_path: Path
) -> Tuple[str, AllTypes]:
    """Checks for a subsection and parses it if found. If a subsection
    is not found it will return arg_name and prop_value unchanged.

    Args:
        arg_name (str): The Terraform argument name.
        prop_value (Any): The Cloudformation property to be converted to Terraform argument.

    Returns:
        Tuple[str, Any]: The arg_name and prop_value.
    """

    section_name = find_section(arg_name, docs_path)

    if not section_name:
        if isinstance(prop_value, dict):
            log.debug(f"{arg_name} has Map value but no subsection in {docs_path}")
            return arg_name, prop_value

        return arg_name, prop_value

    valid_sub_args = doc_file.read_section(docs_path, section_name)

    if not valid_sub_args:
        log.warning(f"{arg_name} has section in {docs_path} but section was empty.")
        return arg_name, prop_value

    log.debug(f"Valid {arg_name} arguments are {valid_sub_args}")

    if not isinstance(prop_value, (dict, list)):
        raise TypeError(
            f"Found section {section_name} but prop_value was {type(prop_value).__name__} not dict or list."
        )

    # todo Nested blocks now have better support but this function will need to be refactored.
    # Currently the function returns a tuple where the key is argument and the value is some terraform type.
    # This works if the section is a key -> map in Cloudformation and nested block in terraform. But
    # in some cases its key -> list in Cloudformation and Block, Block, Block in Terraform. This function
    # will need to be modified so it returns something capable of expressing this relationship.
    if isinstance(prop_value, list):
        sub_args: List[TerraformType] = []

        for sub_props in prop_value:
            if not isinstance(sub_props, Dict):
                sub_args.append(CommentType(sub_props))

            try:
                sub_args.append(props_to_args(sub_props, valid_sub_args, docs_path))
            except:  # noqa: E722
                sub_args.append(CommentType(sub_props))

        return arg_name, ListType(sub_args)

    sub_attrs = props_to_args(prop_value, valid_sub_args, docs_path)
    return arg_name, Block(arg_name, (), sub_attrs)


def convert_parameter_type(param_type: str):
    type_conversion = {
        "String": "string",
        "Number": "number",
        "List<Number>": "list(number)",
        "CommaDelimitedList": "list(string)",
    }

    # todo We might need to handle other types like the SSM ones?
    if param_type not in type_conversion:
        return LiteralType("string")

    return LiteralType(type_conversion[param_type])


def add_space():
    if log.level == logging.DEBUG:
        print()


def perform_resource_overrides(
    tf_type: str, params: Dict[str, TerraformType], tc: TemplateConverter
):
    log.debug(f"Overiding params for {tf_type}")
    if tf_type not in OVERRIDE_DISPATCH:
        return params

    param_overrides = OVERRIDE_DISPATCH[tf_type]

    for param, override in param_overrides.items():
        if param in params:
            params = override(tc, params)

    return params


def perform_global_overrides(
    tf_type: str, params: Dict[str, TerraformType], tc: TemplateConverter
):
    log.debug(f"Performing global overrides for {tf_type}")

    for param, override in GLOBAL_OVERRIDES.items():
        if param in params:
            params = override(tc, params)

    return params


# All the other Cloudformation intrinsic functions start with `Fn:` but for some reason
# the Condition function does not. This can be problem because
#  `Condition` is a valid key in an IAM policy but its value is always a Map.
def is_condition_func(value: Any) -> bool:
    """Checks if the 'Condition' key is a instrinsic function.

    Args:
        value (Any): The value of the 'Condition' key.

    Returns:
        bool: True if we think this `Condition` key is an instrinsic function.
    """
    if isinstance(value, str):
        return True

    return False



from abc import abstractmethod
from typing import Any, Union

try:
    from typing import Protocol
except ImportError:
    # Shouldnt be needed, but just in case
    # Python 3.6 and 3.7
    from typing_extensions import Protocol  # type: ignore

import logging

log = logging.getLogger("cf2tf")


class TerraformType(Protocol):
    """A Terraform value of some type."""

    value: Any

    def __str__(self) -> str:
        return self.render(0)

    @classmethod
    @abstractmethod
    def render(self, indent: int) -> str:
        raise NotImplementedError


class StringType(str, TerraformType):
    """A sequence of Unicode characters representing some text, like "hello"."""

    def __init__(self, value: str) -> None:
        """Default constructor

        Args:
            value (str): The value for this Terraform type.
        """
        super().__init__()

        self.value = value

    def __str__(self) -> str:
        return self.render()

    def render(self, _=0):
        return f'"{self.value}"'


class NumberType(int, TerraformType):
    """A numeric value. The number type can represent both whole numbers like 15 and fractional values like 6.283185."""

    def __init__(self, value: Union[int, float]) -> None:
        """Default constructor

        Args:
            value (str): The value for this Terraform type.
        """
        super().__init__()

        self.value = value

    def __str__(self) -> str:
        return str(self.value)

    def render(self, _=0):
        return self.value


class NullType(int, TerraformType):
    """A value that represents absence or omission. If you set an argument of a resource to null, Terraform behaves as though you had completely omitted it."""

    def __init__(self) -> None:
        """Default constructor

        Args:
            value (str): The value for this Terraform type.
        """
        super().__init__()

        self.value: str = "null"

    def __str__(self) -> str:
        return self.render()

    def render(self, _=0):
        return self.value

    # Needed for unit testing
    def __eq__(self, __x: object) -> bool:
        return __x == self.value


class BooleanType(int, TerraformType):
    """A value that represents a boolean."""

    def __init__(self, value: bool) -> None:
        """Default constructor

        Args:
            value (bool): The value for this Terraform type.
        """
        super().__init__()

        self.value: bool = value

    def __str__(self) -> str:
        return self.render()

    def render(self, _=0):
        return str(self.value).lower()


PrimitiveTypes = Union[StringType, NumberType, NullType, BooleanType]



import logging
import re
from io import TextIOWrapper
from pathlib import Path
from typing import List

log = logging.getLogger("cf2tf")


def parse_attributes(docs_path: Path):
    with open(docs_path) as file:
        try:
            arguments = parse_section("Argument Reference", file)
        except Exception as e:
            raise Exception(f"Unable to find arguments in {file.name}") from e

        try:
            attributes = parse_section("Attribute Reference", file)
        except Exception as e:
            raise Exception(f"Unable to find attributes in {file.name}") from e

    return (arguments, attributes)


def read_section(docs_path: Path, section_name: str):
    items: List[str]
    with open(docs_path) as file:
        items = parse_section(section_name, file)

    return items


def parse_section(name: str, file: TextIOWrapper):
    """
    Parses items from a specific section in a file.

    This function first finds the specified section in the file using the `find_section` function.
    Then, it parses the items in the found section using the `parse_items` function and returns a list of these items.

    Args:
        name (str): The name of the section to parse items from.
        file (TextIOWrapper): The file to parse items from.

    Returns:
        List[str]: A list of items found in the section. If no items are found, an empty list is returned.
    """
    # Throws an exception if the section is not found
    find_section(name, file)

    items = parse_items(file)

    # sometimes we dont find items and thats ok
    if not items:
        log.debug(f"Unable to find items in section {name} of {file.name}")

    return items


def parse_items(file: TextIOWrapper):
    """
    Parses attributes from a section in a file.

    This function reads through a file line by line until it finds a line that starts with a "*" character,
    which indicates an attribute. It then extracts the attribute using a regular expression and adds it to a list.
    The function stops parsing attributes when it encounters an empty line or a line that starts with a "#" character,
    which indicates a new section.

    Args:
        file (TextIOWrapper): The file to parse attributes from.

    Returns:
        List[str]: A list of attributes found in the section. If no attributes are found, an empty list is returned.
    """
    attributes: List[str] = []

    while True:
        # We need the  current postion incase we run into the next section
        cur_pos = file.tell()

        line = file.readline()

        if not line:
            log.debug("We have ran out out attributes to parse")
            break

        # If line starts with whitespace its probably a multiline description
        if line.startswith((" ", "\t")):
            continue

        # These should be the attributes we are after
        if line[0] == "*" or line[0] == "-":
            regex = r"`([\w.*]+)`"

            match = re.search(regex, line)

            if not match:
                log.debug(f"Did not find an item to parse in {line}")
                continue

            attribute = match.group(1)

            attributes.append(attribute)

        # We have reached a new markdown section
        if line[0] == "#":
            file.seek(cur_pos)
            break

    return attributes


def all_sections(docs_path: Path):
    sections: List[str] = []

    with open(docs_path) as file:
        while True:
            line = file.readline()

            # We dont care about the top level sections, only ## and ##
            if line.startswith("##"):
                sections.append(line.strip())

            if not line:
                break

    return sections


def find_section(name: str, file: TextIOWrapper):
    """Searches for a markdown section in a file and returns the position of the section.

    This function reads through a file line by line until it finds a line that starts with a "#" character
    and contains the specified name. It then returns the position of the next line in the file.
    If the end of the file is reached without finding the section, it logs a debug message and raises an exception.

    Args:
        name (str): The name of the section to find
        file (TextIOWrapper): The file to search

    Raises:
        Exception: If the section is not found

    Returns:
        int: The position of the next line after the section
    """
    while True:
        cur_pos = file.tell()

        line = file.readline()

        if line.startswith("#") and name in line:
            return cur_pos + 1

        if not line:
            raise Exception(f"Unable to find section {name} in {file.name}")



import logging
from pathlib import Path
from typing import Optional

import click
import click_log

import cf2tf.save
from cf2tf.cloudformation import Template
from cf2tf.convert import TemplateConverter
from cf2tf.terraform import code

log = logging.getLogger("cf2tf")
click_log.basic_config(log)


@click.command()  # type: ignore
@click.version_option()
@click.option("--output", "-o", type=click.Path(exists=False))
@click_log.simple_verbosity_option(log)
@click.argument("template_path", type=click.Path(exists=True))
def cli(output: Optional[str], template_path: str):
    """Convert Cloudformation template into Terraform.

    Args:
        template_path (str): The path to the cloudformation template
    """

    # Need to take this path and parse the cloudformation file
    tmpl_path = Path(template_path)

    # Where/how we will write the results
    output_writer = cf2tf.save.create_writer(output)

    log.info(f"// Converting {tmpl_path.name} to Terraform!")
    log.debug(f"// Template location is {tmpl_path}")

    cf_template = Template.from_yaml(tmpl_path).template

    # Need to get the code from the repo
    search_manger = code.search_manager()

    # Turn Cloudformation template into a Terraform configuration
    config = TemplateConverter(tmpl_path.stem, cf_template, search_manger).convert()

    # Save this configuration to disc
    config.save(output_writer)


if __name__ == "__main__":
    cli()  # type: ignore



import logging
from typing import Dict, List, Optional, Tuple, Union

from cf2tf.terraform.hcl2.custom import CommentType, LiteralType
from cf2tf.terraform.hcl2.primitive import StringType, TerraformType

log = logging.getLogger("cf2tf")

# BlockLabel = Tuple[Union[str, StringType, Tuple[()]]]

BlockLabel = Union[
    Tuple[()],
    Tuple[
        Union[
            str,
            StringType,
        ],
        ...,
    ],
]

Arguments = Dict[str, Union[TerraformType, "Block"]]


class Block:
    """A block creates a child body that is annotated
    with a block type and zero or more block labels."""

    def __init__(
        self,
        block_type: str,
        labels: Optional[BlockLabel] = None,
        arguments: Optional[Arguments] = None,
        valid_arguments: Optional[List[str]] = None,
        valid_attributes: Optional[List[str]] = None,
    ) -> None:
        self.block_type = block_type
        self.labels = labels if labels else ()
        self.arguments = arguments if arguments else {}
        self.valid_arguments = valid_arguments if valid_arguments else []
        self.valid_attributes = valid_attributes if valid_attributes else []

    def base_ref(self):
        return f"{self.block_type}.{'.'.join(self.labels)}".replace('"', "")

    def __repr__(self) -> str:
        return self.base_ref()

    def __str__(self) -> str:
        return self.render()

    def ref(self, attribute_name: Optional[str] = None):
        count = "[0]" if "count" in self.arguments else ""

        resource_name = f"{self.base_ref()}{count}"

        attribute = (
            self.valid_attributes[0] if attribute_name is None else attribute_name
        )

        return LiteralType(f"{resource_name}.{attribute}")

    def render(self, indent=0):
        brace_space = " " * indent

        indent += 2

        block_labels = " ".join(str(label) for label in self.labels)

        label_space = " " if self.labels else ""

        arg_newline = "\n" if self.arguments else ""

        block_args = render_arguments(self.arguments, indent)

        block = f"""{brace_space}{self.block_type}{label_space}{block_labels} {{{arg_newline}{block_args}{arg_newline}{brace_space}}}"""

        return block


def render_arguments(args: Arguments, indent=0):
    if not args:
        return ""

    indent_spacing = " " * indent

    results = []

    for name, value in args.items():
        try:
            if isinstance(value, (Block, CommentType)):
                results.append(value.render(indent))
                continue

            results.append(f"{indent_spacing}{name} = {value.render(indent)}")
        except AttributeError as ex:
            log.debug(f"Key is type {type(name)} with value {value} ")
            log.debug(f"Value is type {type(value)} with value {value} ")
            raise Exception(
                f"Failed to render argument {name} with value:\n{value}"
            ) from ex

    return "\n".join(results)



from __future__ import annotations

import logging
from pathlib import Path
from typing import Any, Dict, Optional, Union

import yaml  # type: ignore
from cfn_tools import dump_yaml, load_yaml  # type: ignore

log = logging.getLogger("cf2tf")


class Template:
    """Loads a Cloudformation template file so that it's parameters
    and conditions can be rendered into their final form for testing.
    """

    AccountId: str = "5" * 12
    NotificationARNs: list = []
    NoValue: str = ""  # Not yet implemented
    Partition: str = "aws"  # Other regions not implemented
    Region: str = "us-east-1"
    StackId: str = ""  # Not yet implemented
    StackName: str = ""  # Not yet implemented
    URLSuffix: str = "amazonaws.com"  # Other regions not implemented

    def __init__(
        self, template: Dict[str, Any], imports: Optional[Dict[str, str]] = None
    ) -> None:
        """Loads a Cloudformation template from a file and saves
        it as a dictionary.

        Args:
            template (Dict): The Cloudformation template as a dictionary.
            imports (Optional[Dict[str, str]], optional): Values this template plans
            to import from other stacks exports. Defaults to None.

        Raises:
            TypeError: If template is not a dictionary.
            TypeError: If imports is not a dictionary.
        """

        if imports is None:
            imports = {}

        if not isinstance(template, dict):
            raise TypeError(
                f"Template should be a dict, not {type(template).__name__}."
            )

        if not isinstance(imports, dict):
            raise TypeError(f"Imports should be a dict, not {type(imports).__name__}.")

        self.raw: str = yaml.dump(template)
        self.template = template
        self.Region = Template.Region
        self.imports = imports

    @classmethod
    def from_yaml(
        cls, template_path: Union[str, Path], imports: Optional[Dict[str, str]] = None
    ) -> Template:
        """Loads a Cloudformation template from file.

        Args:
            template_path (Union[str, Path]): The path to the template.
            imports (Optional[Dict[str, str]], optional): Values this template plans
            to import from other stacks exports. Defaults to None.

        Returns:
            Template: A Template object ready for testing.
        """

        with open(template_path) as f:
            raw = f.read()

        tmp_yaml = load_yaml(raw)

        tmp_str = dump_yaml(tmp_yaml)

        template = yaml.load(tmp_str, Loader=yaml.FullLoader)

        return cls(template, imports)



"""Save the results of the conversion.
"""

import logging
from itertools import groupby
from pathlib import Path
from typing import Iterable, List, Optional, Type

try:
    from typing import Protocol
except ImportError:
    # Shouldnt be needed, but just in case
    # Python 3.6 and 3.7
    from typing_extensions import Protocol  # type: ignore

import cf2tf.terraform.blocks as hlc2

log = logging.getLogger("cf2tf")


class Output(Protocol):
    def save(self, resources: List[hlc2.Block]) -> None:
        """Save the results

        Args:
            resources (List[hlc2.Block]): The resources to be saved.
        """


class Directory:
    def __init__(self, directory: str) -> None:
        output_dir = Path(directory)

        if not output_dir.exists():
            output_dir.mkdir(parents=True, exist_ok=True)

        if not output_dir.is_dir():
            raise Exception(f"Output {output_dir} should be a directory.")

        self.output_dir = output_dir

    def save(self, resources: List[hlc2.Block]) -> None:
        """Save the results to a directory where each resource type gets it's own file.

        Args:
            resources (List[hlc2.Block]): The resources to be saved.
        """
        log.info(f"Saving converted terraform objects to {self.output_dir.absolute()}")
        for k, g in groupby(resources, lambda s: type(s)):
            self.write_group(k, g)

    def write_group(self, block_type: Type[hlc2.Block], blocks: Iterable[hlc2.Block]):
        """Creates a file and writes the resources into it.

        Args:
            block_type (Type[hlc2.Block]): The type of terraform blocks
            blocks (List[hlc2.Block]): A list of terraform blocks.
        """
        block_name = block_type.__name__.lower()

        file_name = f"{block_name}.tf"

        file_path = self.output_dir / file_name

        with file_path.open("w", encoding="utf-8") as f:
            for block in blocks:
                contents = str(block)
                f.write(contents)
                f.write("\n\n")


class StdOut:
    def save(self, resources: List[hlc2.Block]) -> None:
        """Save the results to stdout

        Args:
            resources (List[hlc2.Block]): The resources to be saved.
        """

        for resource in resources:
            try:
                print()
                print(resource)
            except Exception as e:
                print(f"Unable to write {'.'.join(resource.labels)}")
                raise e


def create_writer(output: Optional[str]) -> Output:
    writer: Output

    if output:
        writer = Directory(output)
    else:
        writer = StdOut()

    return writer



from typing import Union

from cf2tf.terraform.hcl2.primitive import TerraformType


class LiteralType(str, TerraformType):
    """A literal value like the result of a terraform expression."""

    def __init__(self, value: str) -> None:
        super().__init__()
        self.value = value

    def __str__(self) -> str:
        return self.render()

    def render(self, _=0):
        return self.value


class CommentType(str, TerraformType):
    """A comment in the Terraform file."""

    def __init__(self, value: str) -> None:
        super().__init__()
        self.value = value

    def __str__(self) -> str:
        return self.render()

    def render(self, indent=0):
        spacing = " " * indent

        text = [f"{spacing}// {line}" for line in self.value.split("\n")]

        return "\n".join(text)


CustomTypes = Union[LiteralType, CommentType]



import logging
import re
from pathlib import Path
from shutil import rmtree
from tempfile import gettempdir
from typing import Optional

import click
from click._termui_impl import ProgressBar
from git import RemoteProgress
from git.repo.base import InvalidGitRepositoryError, Repo
from thefuzz import fuzz, process  # type: ignore

# import cf2tf.convert

log = logging.getLogger("cf2tf")


class SearchManager:
    def __init__(self, docs_path: Path) -> None:
        self.docs_path = docs_path
        self.resources = list(docs_path.joinpath("r").glob("*.markdown"))
        self.datas = list(docs_path.joinpath("d").glob("*.markdown"))

    def find(self, resource_type: str) -> Path:
        name = resource_type_to_name(resource_type)

        log.debug(f"Searcing for {name} in terraform docs...")

        files = {
            doc_file: transform_file_name(doc_file.name) for doc_file in self.resources
        }

        resource_name: str
        ranking: int
        doc_path: Path
        resource_name, ranking, doc_path = process.extractOne(
            name.lower(), files, scorer=fuzz.ratio
        )

        log.debug(
            f"Best match was {resource_name} at {doc_path} with score of {ranking}."
        )

        return doc_path


def search_manager():
    docs_dir = "website/docs"

    repo = get_code()

    docs_path = Path(repo.working_dir).joinpath(docs_dir)

    if not docs_path.exists():
        print("The docs path does not exist")

    return SearchManager(docs_path)


def get_code() -> Repo:
    temp_dir = Path(gettempdir())
    repo_path = temp_dir.joinpath("terraform_src")

    try:
        existing_repo = repo_from_existing(repo_path)

        if existing_repo:
            return existing_repo

    except InvalidGitRepositoryError:
        rmtree(repo_path)
        repo_path.rmdir()

    print(f"// Cloning Terraform src code to {repo_path}...", end="")

    repo_path.mkdir(exist_ok=True)

    repo = Repo.clone_from(
        "https://github.com/hashicorp/terraform-provider-aws.git",
        repo_path,
        depth=1,
        progress=CloneProgress(),  # type: ignore
    )
    click.echo(" code has been checked out.")

    return repo


def repo_from_existing(repo_path: Path) -> Optional[Repo]:
    if repo_path.exists():
        if repo_path.joinpath(".git").exists():
            repo = Repo(repo_path)
            click.echo(f"// Existing Terraform src code found at {repo_path}.")
            return repo

    return None


def resource_type_to_name(resource_type: str) -> str:
    """Converts a Cloudformation Resource Type into something more search friendly.

    Args:
        resource_type (str): The Cloudformation resource type.

    Returns:
        str: A search term that can be used to match resources in the TF docs.
    """

    search_tokens = resource_type.replace("::", " ").replace("AWS", " ").split(" ")

    # I will leave the logic for camel case splitting here for now.
    # in case we want to use it later.
    # for i, token in enumerate(search_tokens):
    #     if len(token) >= 4:
    #         search_tokens[i] = cf2tf.convert.camel_case_split(token)

    search_term = " ".join(search_tokens).lower().strip()

    log.debug(f"Converted CF type {resource_type} to search term {search_term}.")

    return search_term


class CloneProgress(RemoteProgress):
    def __init__(self):
        super().__init__()
        self.pbar: Optional[ProgressBar] = None

    def update(self, op_code, cur_count, max_count=None, message=""):
        if not self.pbar and max_count:
            self.create_pbar(int(max_count))

        if self.pbar:
            self.pbar.length = int(max_count)
            self.pbar.update(1)

    def create_pbar(self, max_count):
        self.pbar = click.progressbar(length=max_count)


def transform_file_name(og_name: str):
    no_extensions = og_name.split(".")[0]

    no_underscores = no_extensions.replace("_", " ")

    # Sometimes file names have v2 in them.
    split_numbers = re.split(r"(v\d)", no_underscores)

    return " ".join([item.strip() for item in split_numbers])



from __future__ import annotations

import logging
from typing import List

from cf2tf.save import Output
from cf2tf.terraform.blocks import Block

log = logging.getLogger("cf2tf")


class Configuration:
    def __init__(self, resources: List[Block]) -> None:
        self.resources = resources

    def save(self, output: Output):
        output.save(self.resources)



from typing import Union

from cf2tf.terraform.hcl2._block import Block
from cf2tf.terraform.hcl2.complex import ComplexTypes
from cf2tf.terraform.hcl2.custom import CustomTypes
from cf2tf.terraform.hcl2.primitive import PrimitiveTypes

AllTypes = Union[Block, PrimitiveTypes, ComplexTypes, CustomTypes]



from typing import TYPE_CHECKING, Callable, Dict

from cf2tf.terraform.hcl2.complex import ListType, MapType
from cf2tf.terraform.hcl2.custom import LiteralType
from cf2tf.terraform.hcl2.primitive import NullType, StringType, TerraformType

if TYPE_CHECKING:
    from cf2tf.convert import TemplateConverter

CFParams = Dict[str, TerraformType]

Override = Callable[["TemplateConverter", CFParams], CFParams]

ParamOverride = Dict[str, Override]

ResourceOverride = Dict[str, ParamOverride]


def s3_bucket_acl(_tc: "TemplateConverter", params: CFParams) -> CFParams:
    access_controls: CFParams = {
        "Private": StringType("private"),
        "PublicRead": StringType("public-read"),
        "PublicReadWrite": StringType("public-read-write"),
        "AuthenticatedRead": StringType("authenticated-read"),
        "LogDeliveryWrite": StringType("log-delivery-write"),
        "BucketOwnerRead": NullType(),
        "BucketOwnerFullControl": NullType(),
        "AwsExecRead": StringType("aws-exec-read"),
    }

    orig_value = params["AccessControl"]

    new_value = access_controls[orig_value]  # type: ignore

    del params["AccessControl"]

    params["acl"] = new_value
    return params


def s3_bucket_policy(_tc: "TemplateConverter", params: CFParams) -> CFParams:
    policy_value = params["PolicyDocument"]

    params["PolicyDocument"] = LiteralType(f"jsonencode({policy_value.render(4)}\n  )")

    return params


def tag_conversion(_tc: "TemplateConverter", params: CFParams) -> CFParams:
    if isinstance(params["Tags"], dict):
        return params

    original_tags: ListType = params["Tags"]  # type: ignore

    first_item = original_tags[0]

    # It's possible that the tags might be one or more
    # conditional statements of LiteralType
    # This wont fix every case, but it should fix most
    # and it's better than nothing
    if not isinstance(first_item, (dict, MapType)):
        del params["Tags"]
        params["tags"] = first_item
        return params

    try:
        new_tags = {LiteralType(tag["Key"]): tag["Value"] for tag in original_tags}

        del params["Tags"]
        params["tags"] = MapType(new_tags)
        return params
    except Exception:
        del params["Tags"]
        params["tags"] = LiteralType(
            f"// Could not convert tags: {original_tags.render(4)}"
        )
        return params

    raise Exception("Could not convert tags")


OVERRIDE_DISPATCH: ResourceOverride = {
    "aws_s3_bucket": {"AccessControl": s3_bucket_acl},
    "aws_s3_bucket_policy": {"PolicyDocument": s3_bucket_policy},
}

GLOBAL_OVERRIDES: ParamOverride = {"Tags": tag_conversion}



import logging
from typing import Dict, List, Union

from cf2tf.terraform.hcl2.primitive import PrimitiveTypes, TerraformType

log = logging.getLogger("cf2tf")


class ListType(list, TerraformType):
    def __init__(self, value: List[TerraformType]) -> None:
        """Default constructor

        Args:
            value (str): The value for this Terraform type.
        """
        super().__init__(value)
        self.value = value

    def __str__(self) -> str:
        return self.render()

    def render(self, indent=0):
        return render_tf_list(self, indent)


class MapType(dict, TerraformType):
    def __init__(self, value=Dict[PrimitiveTypes, TerraformType]) -> None:
        """Default constructor

        Args:
            value (str): The value for this Terraform type.
        """
        super().__init__(value)
        self.value = value

    def __str__(self) -> str:
        return self.render()

    def render(self, indent=0):
        return render_tf_map(self, indent)


ComplexTypes = Union[ListType, MapType]


def render_tf_list(items: List[TerraformType], indent=0):
    rear_brace = " " * indent

    indent += 2

    spacing = " " * indent

    result = "[\n"

    for item in items:
        comma = "" if item is items[-1] else ","

        result += f"{spacing}{item.render(indent)}{comma}\n"

    result += f"{rear_brace}]"

    return result


def render_tf_map(items: Dict[PrimitiveTypes, TerraformType], indent=0):
    rear_brace = " " * indent

    indent += 2

    spacing = " " * indent

    result = "{"

    for name, value in items.items():
        # if isinstance(value, ListType):
        #     result += f"\n{spacing}{name} = {value.render(indent)}"

        result = result + f"\n{spacing}{name} = {value.render(indent)}"

    return result + f"\n{rear_brace}}}"
